@@ -20,6 +20,7 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
+#include <assert.h>
 #include <stdio.h>
 #include <errno.h>
 #include <limits.h>
@@ -27,6 +28,8 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <string.h>
+#include <pthread.h>
+#include <time.h>
 
 #define __USE_GNU /* for *_CLOEXEC */
 
@@ -37,30 +40,15 @@
 #include "caps.h"
 #include "utils.h"
 
-#define LXC_LOG_PREFIX_SIZE	32
-#define LXC_LOG_BUFFER_SIZE	512
+#define LXC_LOG_DATEFOMAT_SIZE  15
 
-#ifdef HAVE_TLS
-__thread int lxc_log_fd = -1;
-static __thread char log_prefix[LXC_LOG_PREFIX_SIZE] = "lxc";
-static __thread char *log_fname = NULL;
-/* command line values for logfile or logpriority should always override
- * values from the configuration file or defaults
- */
-static __thread int lxc_logfile_specified = 0;
-static __thread int lxc_loglevel_specified = 0;
-static __thread int lxc_quiet_specified = 0;
-#else
 int lxc_log_fd = -1;
+int lxc_quiet_specified;
+int lxc_log_use_global_fd;
+static int lxc_loglevel_specified;
+
 static char log_prefix[LXC_LOG_PREFIX_SIZE] = "lxc";
 static char *log_fname = NULL;
-static int lxc_quiet_specified = 0;
-/* command line values for logfile or logpriority should always override
- * values from the configuration file or defaults
- */
-static int lxc_logfile_specified = 0;
-static int lxc_loglevel_specified = 0;
-#endif
 
 lxc_log_define(lxc_log, lxc);
 
@@ -72,6 +60,7 @@ static int log_append_stderr(const struct lxc_log_appender *appender,
 		return 0;
 
 	fprintf(stderr, "%s: ", log_prefix);
+	fprintf(stderr, "%s: %s: %d ", event->locinfo->file, event->locinfo->func, event->locinfo->line);
 	vfprintf(stderr, event->fmt, *event->vap);
 	fprintf(stderr, "\n");
 	return 0;
@@ -81,21 +70,36 @@ static int log_append_stderr(const struct lxc_log_appender *appender,
 static int log_append_logfile(const struct lxc_log_appender *appender,
 			      struct lxc_log_event *event)
 {
+	char date[LXC_LOG_DATEFOMAT_SIZE] = "20150427012246";
 	char buffer[LXC_LOG_BUFFER_SIZE];
+	const struct tm *t;
 	int n;
 	int ms;
+	int fd_to_use = -1;
 
-	if (lxc_log_fd == -1)
+#ifndef NO_LXC_CONF
+	if (!lxc_log_use_global_fd && current_config)
+		fd_to_use = current_config->logfd;
+#endif
+
+	if (fd_to_use == -1)
+		fd_to_use = lxc_log_fd;
+
+	if (fd_to_use == -1)
 		return 0;
 
+	t = localtime(&event->timestamp.tv_sec);
+	strftime(date, sizeof(date), "%Y%m%d%H%M%S", t);
 	ms = event->timestamp.tv_usec / 1000;
 	n = snprintf(buffer, sizeof(buffer),
-		     "%15s %10ld.%03d %-8s %s - ",
+		     "%15s %10s.%03d %-8s %s - %s:%s:%d - ",
 		     log_prefix,
-		     event->timestamp.tv_sec,
+		     date,
 		     ms,
 		     lxc_log_priority_to_string(event->priority),
-		     event->category);
+		     event->category,
+		     event->locinfo->file, event->locinfo->func,
+		     event->locinfo->line);
 
 	n += vsnprintf(buffer + n, sizeof(buffer) - n, event->fmt,
 		       *event->vap);
@@ -108,7 +112,7 @@ static int log_append_logfile(const struct lxc_log_appender *appender,
 
 	buffer[n] = '\n';
 
-	return write(lxc_log_fd, buffer, n + 1);
+	return write(fd_to_use, buffer, n + 1);
 }
 
 static struct lxc_log_appender log_appender_stderr = {
@@ -133,7 +137,7 @@ static struct lxc_log_category log_root = {
 struct lxc_log_category lxc_log_category_lxc = {
 	.name		= "lxc",
 	.priority	= LXC_LOG_PRIORITY_ERROR,
-	.appender	= &log_appender_stderr,
+	.appender	= &log_appender_logfile,
 	.parent		= &log_root
 };
 
@@ -156,7 +160,7 @@ static int build_dir(const char *name)
 		*p = '\0';
 		if (access(n, F_OK)) {
 			ret = lxc_unpriv(mkdir(n, 0755));
-			if (ret && errno != -EEXIST) {
+			if (ret && errno != EEXIST) {
 				SYSERROR("failed to create directory '%s'.", n);
 				free(n);
 				return -1;
@@ -197,7 +201,7 @@ static int log_open(const char *name)
  * Build the path to the log file
  * @name     : the name of the container
  * @lxcpath  : the lxcpath to use as a basename or NULL to use LOGPATH
- * Returns malloced path on sucess, or NULL on failure
+ * Returns malloced path on success, or NULL on failure
  */
 static char *build_log_path(const char *name, const char *lxcpath)
 {
@@ -247,6 +251,16 @@ static char *build_log_path(const char *name, const char *lxcpath)
 	return p;
 }
 
+extern void lxc_log_close(void)
+{
+	if (lxc_log_fd == -1)
+		return;
+	close(lxc_log_fd);
+	lxc_log_fd = -1;
+	free(log_fname);
+	log_fname = NULL;
+}
+
 /*
  * This can be called:
  *   1. when a program calls lxc_log_init with no logfile parameter (in which
@@ -259,11 +273,12 @@ static int __lxc_log_set_file(const char *fname, int create_dirs)
 {
 	if (lxc_log_fd != -1) {
 		// we are overriding the default.
-		close(lxc_log_fd);
-		free(log_fname);
+		lxc_log_close();
 	}
 
-	if (!fname || strlen(fname) == 0) {
+	assert(fname != NULL);
+
+	if (strlen(fname) == 0) {
 		log_fname = NULL;
 		return 0;
 	}
@@ -302,6 +317,11 @@ static int _lxc_log_set_file(const char *name, const char *lxcpath, int create_d
 	return ret;
 }
 
+/*
+ * lxc_log_init:
+ * Called from lxc front-end programs (like lxc-create, lxc-start) to
+ * initalize the log defaults.
+ */
 extern int lxc_log_init(const char *name, const char *file,
 			const char *priority, const char *prefix, int quiet,
 			const char *lxcpath)
@@ -317,10 +337,12 @@ extern int lxc_log_init(const char *name, const char *file,
 	if (priority)
 		lxc_priority = lxc_log_priority_to_int(priority);
 
-	lxc_log_category_lxc.priority = lxc_priority;
+	if (!lxc_loglevel_specified) {
+		lxc_log_category_lxc.priority = lxc_priority;
+		lxc_loglevel_specified = 1;
+	}
 
 	if (!lxc_quiet_specified) {
-		lxc_log_category_lxc.appender = &log_appender_logfile;
 		if (!quiet)
 			lxc_log_category_lxc.appender->next = &log_appender_stderr;
 	}
@@ -332,12 +354,8 @@ extern int lxc_log_init(const char *name, const char *file,
 		if (strcmp(file, "none") == 0)
 			return 0;
 		ret = __lxc_log_set_file(file, 1);
+		lxc_log_use_global_fd = 1;
 	} else {
-
-		/* For now, unprivileged containers have to set -l to get logging */
-		if (geteuid())
-			return 0;
-
 		/* if no name was specified, there nothing to do */
 		if (!name)
 			return 0;
@@ -347,8 +365,8 @@ extern int lxc_log_init(const char *name, const char *file,
 		if (!lxcpath)
 			lxcpath = LOGPATH;
 
-		/* try LOGPATH if lxcpath is the default */
-		if (strcmp(lxcpath, lxc_global_config_value("lxc.lxcpath")) == 0)
+		/* try LOGPATH if lxcpath is the default for the privileged containers */
+		if (!geteuid() && strcmp(LXCPATH, lxcpath) == 0)
 			ret = _lxc_log_set_file(name, NULL, 0);
 
 		/* try in lxcpath */
@@ -372,30 +390,18 @@ extern int lxc_log_init(const char *name, const char *file,
 	return ret;
 }
 
-extern void lxc_log_close(void)
-{
-	if (lxc_log_fd == -1)
-		return;
-	close(lxc_log_fd);
-	lxc_log_fd = -1;
-	free(log_fname);
-	log_fname = NULL;
-}
-
 /*
  * This is called when we read a lxc.loglevel entry in a lxc.conf file.  This
  * happens after processing command line arguments, which override the .conf
  * settings.  So only set the level if previously unset.
  */
-extern int lxc_log_set_level(int level)
+extern int lxc_log_set_level(int *dest, int level)
 {
-	if (lxc_loglevel_specified)
-		return 0;
 	if (level < 0 || level >= LXC_LOG_PRIORITY_NOTSET) {
 		ERROR("invalid log priority %d", level);
 		return -1;
 	}
-	lxc_log_category_lxc.priority = level;
+	*dest = level;
 	return 0;
 }
 
@@ -417,11 +423,23 @@ extern bool lxc_log_has_valid_level(void)
  * happens after processing command line arguments, which override the .conf
  * settings.  So only set the file if previously unset.
  */
-extern int lxc_log_set_file(const char *fname)
+extern int lxc_log_set_file(int *fd, const char *fname)
 {
-	if (lxc_logfile_specified)
-		return 0;
-	return __lxc_log_set_file(fname, 0);
+	if (*fd != -1) {
+		close(*fd);
+		*fd = -1;
+	}
+
+	if (build_dir(fname)) {
+		ERROR("failed to create dir for log file \"%s\" : %s", fname,
+				strerror(errno));
+		return -1;
+	}
+
+	*fd = log_open(fname);
+	if (*fd == -1)
+		return -errno;
+	return 0;
 }
 
 extern const char *lxc_log_get_file(void)
@@ -442,11 +460,6 @@ extern const char *lxc_log_get_prefix(void)
 
 extern void lxc_log_options_no_override()
 {
-	if (lxc_log_get_file())
-		lxc_logfile_specified = 1;
-
-	if (lxc_log_get_level() != LXC_LOG_PRIORITY_NOTSET)
-		lxc_loglevel_specified = 1;
-
 	lxc_quiet_specified = 1;
+	lxc_loglevel_specified = 1;
 }
