@@ -117,6 +117,12 @@ struct lxc_attach_python_payload {
 
 static int lxc_attach_python_exec(void* _payload)
 {
+    /* This function is the first one to be called after attaching to a
+     * container. As lxc_attach() calls fork() PyOS_AfterFork should be called
+     * in the new process if the Python interpreter will continue to be used.
+     */
+    PyOS_AfterFork();
+
     struct lxc_attach_python_payload *payload =
         (struct lxc_attach_python_payload *)_payload;
     PyObject *result = PyObject_CallFunctionObjArgs(payload->fn,
@@ -237,8 +243,7 @@ void lxc_attach_free_options(lxc_attach_options_t *options)
     int i;
     if (!options)
         return;
-    if (options->initial_cwd)
-        free(options->initial_cwd);
+    free(options->initial_cwd);
     if (options->extra_env_vars) {
         for (i = 0; options->extra_env_vars[i]; i++)
             free(options->extra_env_vars[i]);
@@ -444,7 +449,9 @@ Container_init(Container *self, PyObject *args, PyObject *kwds)
     self->container = lxc_container_new(name, config_path);
     if (!self->container) {
         Py_XDECREF(fs_config_path);
-        fprintf(stderr, "%d: error creating container %s\n", __LINE__, name);
+
+        PyErr_Format(PyExc_RuntimeError, "%s:%s:%d: error during init for container '%s'.",
+			__FUNCTION__, __FILE__, __LINE__, name);
         return -1;
     }
 
@@ -519,6 +526,66 @@ Container_state(Container *self, void *closure)
 }
 
 /* Container Functions */
+static PyObject *
+Container_attach_interface(Container *self, PyObject *args, PyObject *kwds)
+{
+    static char *kwlist[] = {"src_ifname", "dst_ifname", NULL};
+    char *src_name = NULL;
+    char *dst_name = NULL;
+    PyObject *py_src_name = NULL;
+    PyObject *py_dst_name = NULL;
+
+    if (! PyArg_ParseTupleAndKeywords(args, kwds, "O&|O&", kwlist,
+                                      PyUnicode_FSConverter, &py_src_name,
+                                      PyUnicode_FSConverter, &py_dst_name))
+        return NULL;
+
+    if (py_src_name != NULL) {
+        src_name = PyBytes_AS_STRING(py_src_name);
+        assert(src_name != NULL);
+    }
+
+    if (py_dst_name != NULL) {
+        dst_name = PyBytes_AS_STRING(py_dst_name);
+        assert(dst_name != NULL);
+    }
+
+    if (self->container->attach_interface(self->container, src_name, dst_name)) {
+        Py_XDECREF(py_src_name);
+        Py_XDECREF(py_dst_name);
+        Py_RETURN_TRUE;
+    }
+
+    Py_XDECREF(py_src_name);
+    Py_XDECREF(py_dst_name);
+    Py_RETURN_FALSE;
+}
+
+static PyObject *
+Container_detach_interface(Container *self, PyObject *args, PyObject *kwds)
+{
+    static char *kwlist[] = {"ifname", NULL};
+    char *ifname = NULL;
+    PyObject *py_ifname = NULL;
+
+    if (! PyArg_ParseTupleAndKeywords(args, kwds, "O&", kwlist,
+                                      PyUnicode_FSConverter, &py_ifname))
+        return NULL;
+
+    if (py_ifname != NULL) {
+        ifname = PyBytes_AS_STRING(py_ifname);
+        assert(ifname != NULL);
+    }
+
+    if (self->container->detach_interface(self->container, ifname, NULL)) {
+        Py_XDECREF(py_ifname);
+        Py_RETURN_TRUE;
+    }
+
+    Py_XDECREF(py_ifname);
+    Py_RETURN_FALSE;
+}
+
 static PyObject *
 Container_add_device_node(Container *self, PyObject *args, PyObject *kwds)
 {
@@ -741,11 +808,11 @@ Container_create(Container *self, PyObject *args, PyObject *kwds)
     char** create_args = {NULL};
     PyObject *retval = NULL;
     PyObject *vargs = NULL;
+    char *bdevtype = NULL;
     int i = 0;
-    static char *kwlist[] = {"template", "flags", "args", NULL};
-
-    if (! PyArg_ParseTupleAndKeywords(args, kwds, "|siO", kwlist,
-                                      &template_name, &flags, &vargs))
+    static char *kwlist[] = {"template", "flags", "bdevtype", "args", NULL};
+    if (! PyArg_ParseTupleAndKeywords(args, kwds, "|sisO", kwlist,
+                                      &template_name, &flags, &bdevtype, &vargs))
         return NULL;
 
     if (vargs) {
@@ -761,7 +828,7 @@ Container_create(Container *self, PyObject *args, PyObject *kwds)
         }
     }
 
-    if (self->container->create(self->container, template_name, NULL, NULL,
+    if (self->container->create(self->container, template_name, bdevtype, NULL,
                                 flags, create_args))
         retval = Py_True;
     else
@@ -1470,6 +1537,18 @@ static PyGetSetDef Container_getseters[] = {
 };
 
 static PyMethodDef Container_methods[] = {
+    {"attach_interface", (PyCFunction)Container_attach_interface,
+     METH_VARARGS|METH_KEYWORDS,
+     "attach_interface(src_ifname, dest_ifname) -> boolean\n"
+     "\n"
+     "Pass a new network device to the container."
+    },
+    {"detach_interface", (PyCFunction)Container_detach_interface,
+     METH_VARARGS|METH_KEYWORDS,
+     "detach_interface(ifname) -> boolean\n"
+     "\n"
+     "detach a network device from the container."
+    },
     {"add_device_node", (PyCFunction)Container_add_device_node,
      METH_VARARGS|METH_KEYWORDS,
      "add_device_node(src_path, dest_path) -> boolean\n"
